import json
from typing import Any

import numpy as np

from ..joint import Joint
from ..link import Link
from ..matrix import Matrix


def load_model(path: str) -> tuple[Link, dict[str, Joint] | None]:
    model = _load_model(path)

    root = _load_link(model["root"])

    joints = _load_joints(model["joints"], root) if "joints" in model else None

    if "pose" in model:
        _load_pose(model["pose"], joints)

    return root, joints


def load_pose(path: str, joints: dict[str, Joint]) -> None:
    model = _load_model(path)

    _load_pose(model["pose"], joints)


def _load_model(path: str) -> dict[str, Any]:
    try:
        with open(path) as file:
            model = json.load(file)
    except Exception as e:
        raise ValueError(f"Unable to load {path}. {e}")

    if not isinstance(model, dict):
        raise ValueError(f"{path} is not a JSON object")

    return model


def _load_link(value: dict[str, Any]) -> Link:
    root = Link(
        value["name"],
        _matrix(value["end"]),
        _matrix(value["wheel_axis"]) if "wheel_axis" in value else None,
    )

    for link, matrix in value["links"]:
        root.connect(_load_link(link), _matrix(matrix))

    return root


def _matrix(value: list) -> Matrix:
    return Matrix(np.array(value))


def _load_joints(value: dict[str, Any], root: Link) -> dict[str, Joint]:
    links = {l.name: l for l in root.flatten()}

    return {
        n: Joint(
            links[j["link"]],
            _matrix(j["alignment"]),
            j["alpha_limit"],
            j["beta_limit"],
            j["gamma_limit"],
        )
        for n, j in value.items()
    }


def _load_pose(value: dict[str, Any], joints: dict[str, Joint]) -> None:
    pose = {n: (a["alpha"], a["beta"], a["gamma"]) for n, a in value.items()}

    for name, angles in pose.items():
        joints[name].set(*angles)



import socket

from .primatives import Primitive


class Connection:
    def __init__(self, ip_address: str = "localhost", port: int = 6000) -> None:
        self.__address = (ip_address, port)

        self.__socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.__socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 65535)

        self.__buffer_size = self.__socket.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)

    def __del__(self) -> None:
        self.__socket.close()

    def send(self, primitives: list[Primitive]) -> None:
        json = "[" + ",".join([str(p) for p in primitives]) + "]"

        data = json.encode("ascii")

        if len(data) > self.__buffer_size:
            raise ValueError(f"The data size is {len(data)}, which exceeds the buffer size of {self.__buffer_size}.")

        self.__socket.sendto(data, self.__address)



import numpy as np

from ..link import Link
from ..matrix import Matrix

# Translates the root link relative to the world.


def translate(root: Link, xyz: np.ndarray) -> None:
    if not root.is_root:
        raise ValueError(f"{root.name} is not the root")

    root.joint = Matrix(xyz=xyz, rotation=root.joint.rotation)



from enum import Enum

from ..link import Link
from ..matrix import Matrix

# Calibrates the alignment of all IMUs by comparing a predefined calibration
# pose with the IMU measurements while the subject replicates the pose.
#
# Mounting describes the orientation of the IMU attached to the root link. If
# mounting is specified then this IMU must be physically aligned so that one of
# its principal axes points precisely forwards or backwards. This axis may be
# inclined (e.g. tilted up or down) but must not have any component pointing
# left or right. If mounting is not specified then the heading of the
# replicated pose must match the heading of the predefined calibration pose.


class Mounting(Enum):
    X_FORWARDS = Matrix.align_px_py_pz()
    X_BACKWARDS = Matrix.align_nx_ny_pz()
    Y_FORWARDS = Matrix.align_py_nx_pz()
    Y_BACKWARDS = Matrix.align_ny_px_pz()
    Z_FORWARDS = Matrix.align_pz_ny_px()
    Z_BACKWARDS = Matrix.align_nz_py_px()


def calibrate(
    root: Link,
    imus: dict[str, Matrix],  # {<link name>: <IMU measurment>, ...}
    pose: dict[str, Matrix] = {},  # {<link name>: <link joint matrix>, ...}
    mounting: Mounting | None = None,
) -> float | None:
    if not root.is_root:
        raise ValueError(f"{root.name} is not the root")

    links = {l.name: l for l in root.flatten()}

    for link in links.values():
        link.joint = Matrix()

    for name, matrix in pose.items():
        links[name].joint = matrix

    if mounting:
        heading = (imus[root.name] * mounting.value).rot_xyz[2]

        root.joint = Matrix(rotation=(Matrix(rot_z=heading) * root.joint).rotation, xyz=root.joint.xyz)

    for name, matrix in imus.items():
        links[name].set_imu_world(matrix)

    if mounting:
        return heading



from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np

from ..joint import Joint
from ..link import Link
from ..matrix import Matrix


class Primitive(ABC):
    @abstractmethod
    def __str__(self) -> str:
        pass


@dataclass(frozen=True)
class Line(Primitive):
    start: np.ndarray
    end: np.ndarray

    def __str__(self) -> str:
        return f'{{"type":"line","start":{_xyz(self.start)},"end":{_xyz(self.end)}}}'


@dataclass(frozen=True)
class Circle(Primitive):
    xyz: np.ndarray
    axis: np.ndarray
    radius: float

    def __str__(self) -> str:
        return f'{{"type":"circle","xyz":{_xyz(self.xyz)},"axis":{_xyz(self.axis)},"radius":{_number(self.radius)}}}'


@dataclass(frozen=True)
class Dot(Primitive):
    xyz: np.ndarray
    size: float = 1.0

    def __str__(self) -> str:
        return f'{{"type":"dot","xyz":{_xyz(self.xyz)},"size":{_number(self.size)}}}'


@dataclass(frozen=True)
class Axes(Primitive):
    matrix: Matrix
    scale: float = 1.0

    def __str__(self) -> str:
        return f'{{"type":"axes","xyz":{_xyz(self.matrix.xyz)},"quaternion":{_quaternion(self.matrix.quaternion)},"scale":{_number(self.scale)}}}'


@dataclass(frozen=True)
class Angle(Primitive):
    matrix: Matrix
    angle: float
    limit: tuple[float, float] | None = None
    scale: float = 1.0

    def __str__(self) -> str:
        key_values = [
            '"type":"angle"',
            f'"xyz":{_xyz(self.matrix.xyz)}',
            f'"quaternion":{_quaternion(self.matrix.quaternion)}',
            f'"angle":{_number(self.angle)}',
            f'"limit":[{_number(self.limit[0])},{_number(self.limit[1])}]' if self.limit else None,
            f'"scale":{_number(self.scale)}',
        ]

        return "{" + ",".join([k for k in key_values if k]) + "}"


@dataclass(frozen=True)
class Angles(Primitive):
    matrix: Matrix
    alpha: float | None = None
    beta: float | None = None
    gamma: float | None = None
    alpha_limit: tuple[float, float] | None = None
    beta_limit: tuple[float, float] | None = None
    gamma_limit: tuple[float, float] | None = None
    scale: float = 1.0
    mirror: bool = False

    def __str__(self) -> str:
        key_values = [
            '"type":"angles"',
            f'"xyz":{_xyz(self.matrix.xyz)}',
            f'"quaternion":{_quaternion(self.matrix.quaternion)}',
            f'"alpha":{_number(self.alpha)}' if self.alpha is not None else None,
            f'"beta":{_number(self.beta)}' if self.beta is not None else None,
            f'"gamma":{_number(self.gamma)}' if self.gamma is not None else None,
            f'"alpha_limit":[{_number(self.alpha_limit[0])},{_number(self.alpha_limit[1])}]' if self.alpha_limit else None,
            f'"beta_limit":[{_number(self.beta_limit[0])},{_number(self.beta_limit[1])}]' if self.beta_limit else None,
            f'"gamma_limit":[{_number(self.gamma_limit[0])},{_number(self.gamma_limit[1])}]' if self.gamma_limit else None,
            f'"scale":{_number(self.scale)}',
            '"mirror":true' if self.mirror else None,
        ]

        return "{" + ",".join([k for k in key_values if k]) + "}"


@dataclass(frozen=True)
class Label(Primitive):
    xyz: np.ndarray
    text: str

    def __str__(self) -> str:
        return f'{{"type":"label","xyz":{_xyz(self.xyz)},"text":"{self.text}"}}'


@dataclass(frozen=True)
class Pedestal(Primitive):
    xyz: np.ndarray

    def __str__(self) -> str:
        return f'{{"type":"pedestal","xyz":{_xyz(self.xyz)}}}'


def _number(value: float) -> str:
    string = f"{value:.6f}".rstrip("0").rstrip(".")

    return "0" if string == "-0" else string


def _xyz(xyz: np.ndarray) -> str:
    return f"[{_number(xyz[0])},{_number(xyz[1])},{_number(xyz[2])}]"


def _quaternion(quaternion: np.ndarray) -> str:
    return f"[{_number(quaternion[0])},{_number(quaternion[1])},{_number(quaternion[2])},{_number(quaternion[3])}]"


def link_to_primitives(root: Link) -> list[Primitive]:
    primitives = [Pedestal(root.get_joint_world().xyz)]

    for link in root.flatten():
        joint = link.get_joint_world()
        end = link.get_end_world()

        primitives.append(Line(joint.xyz, end.xyz))
        primitives.append(Dot(joint.xyz))
        primitives.append(Axes(joint, 0.5 * link.length))

        imu = link.get_imu_world()

        primitives.append(Dot(imu.xyz, 0.5))
        primitives.append(Axes(imu, 0.25 * link.length))
        primitives.append(Label(imu.xyz, link.name))

        for next_link, _ in link.links:
            next_joint = next_link.get_joint_world()

            primitives.append(Line(joint.xyz, next_joint.xyz))
            primitives.append(Line(end.xyz, next_joint.xyz))

        wheel_axis = link.get_wheel_axis_world()

        if wheel_axis:
            primitives.append(Circle(joint.xyz, wheel_axis.xyz, link.length))

    return primitives


def joints_to_primitives(
    joints: dict[str, Joint],
    mirror: str | None = None,  # angles are mirrored if the joint name contains this string
) -> list[Primitive]:
    primitives = []

    for name, joint in joints.items():
        alpha, beta, gamma = joint.get()

        alpha_limit = joint.alpha_limit
        beta_limit = joint.beta_limit
        gamma_limit = joint.gamma_limit

        joint_world = Matrix(xyz=joint.link.get_joint_world().xyz, rotation=joint.link.origin.rotation)

        primitives.append(
            Angles(
                joint_world * joint.alignment,
                gamma=None if gamma_limit == (0, 0) else gamma,
                beta=None if beta_limit == (0, 0) else beta,
                alpha=None if alpha_limit == (0, 0) else alpha,
                gamma_limit=None if gamma_limit == (0, 0) else gamma_limit,
                beta_limit=None if beta_limit == (0, 0) else beta_limit,
                alpha_limit=None if alpha_limit == (0, 0) else alpha_limit,
                scale=joint.link.length / 3,
                mirror=mirror in name if mirror else False,
            )
        )

        primitives.append(Label(joint_world.xyz, name))

    return primitives
